@@ -47,6 +47,13 @@ from async_timeout import timeout
 import attr
 from cachetools import cached, LRUCache
 import snappy
+from tenacity import (
+    AsyncRetrying,
+    stop_after_attempt,
+    stop_after_delay,
+    retry_if_exception_type,
+    wait_fixed,
+)
 import time
 import zmq, zmq.asyncio
 
@@ -71,11 +78,9 @@ from ai.backend.common.types import (
     KernelCreationResult,
     MountTypes,
     MountPermission,
-    MountTuple5,
-    MountTuple4,
-    MountTuple3,
     Sentinel,
     ServicePortProtocols,
+    VFolderMount,
     aobject,
 )
 from ai.backend.common.events import (
@@ -103,8 +108,7 @@ from ai.backend.common.utils import cancel_tasks, current_loop
 from ai.backend.common.plugin.monitor import ErrorPluginContext, StatsPluginContext
 from ai.backend.common.service_ports import parse_service_ports
 from . import __version__ as VERSION
-from .defs import ipc_base_path
-from .exception import ResourceError
+from .exception import AgentError, ResourceError
 from .kernel import (
     AbstractKernel,
     KernelFeatures,
@@ -182,7 +186,8 @@ class AbstractKernelCreationContext(aobject, Generic[KernelObjectType]):
         self.kernel_config = kernel_config
         self.image_ref = ImageRef(
             kernel_config['image']['canonical'],
-            [kernel_config['image']['registry']['name']],
+            known_registries=[kernel_config['image']['registry']['name']],
+            architecture=kernel_config['image'].get('architecture', get_arch_name()),
         )
         self.internal_data = kernel_config['internal_data'] or {}
         self.computers = computers
@@ -243,7 +248,6 @@ class AbstractKernelCreationContext(aobject, Generic[KernelObjectType]):
         src: Union[str, Path],
         target: Union[str, Path],
         perm: Literal['ro', 'rw'] = 'ro',
-        is_unmanaged: bool = False,
         opts: Mapping[str, Any] = None,
     ):
         """
@@ -297,62 +301,21 @@ class AbstractKernelCreationContext(aobject, Generic[KernelObjectType]):
 
     async def mount_vfolders(
         self,
-        vfolders,
+        vfolders: Sequence[VFolderMount],
         resource_spec: KernelResourceSpec,
     ) -> None:
-        vfolder_mount_map: Mapping[str, str]
-        vfolder_mount_map = self.kernel_config.get('mount_map', {})
         for vfolder in vfolders:
-            is_unmanaged = False
-            # TODO: update to use storage-proxy-provided mount path
-            if len(vfolder) == 5:
-                folder_name, folder_host, folder_id, folder_perm_literal, host_path_raw = \
-                    cast(MountTuple5, vfolder)
-                if host_path_raw:
-                    is_unmanaged = True
-                    host_path = Path(host_path_raw)
-                else:
-                    mount_path = Path(folder_id)
-                    if mount_path.is_absolute():
-                        # Use the storage proxy-provided path as-is.
-                        host_path = mount_path
-                    else:
-                        host_path = (self.local_config['vfolder']['mount'] / folder_host /
-                                     self.local_config['vfolder']['fsprefix'] / folder_id)
-            elif len(vfolder) == 4:  # for backward compatibility
-                folder_name, folder_host, folder_id, folder_perm_literal = \
-                    cast(MountTuple4, vfolder)
-                host_path = (self.local_config['vfolder']['mount'] / folder_host /
-                             self.local_config['vfolder']['fsprefix'] / folder_id)
-            elif len(vfolder) == 3:  # legacy managers
-                folder_name, folder_host, folder_id = cast(MountTuple3, vfolder)
-                folder_perm_literal = 'rw'
-                host_path = (self.local_config['vfolder']['mount'] / folder_host /
-                             self.local_config['vfolder']['fsprefix'] / folder_id)
-            else:
-                raise RuntimeError(
-                    'Unexpected number of vfolder mount detail tuple size')
             if self.internal_data.get('prevent_vfolder_mounts', False):
                 # Only allow mount of ".logs" directory to prevent expose
                 # internal-only information, such as Docker credentials to user's ".docker" vfolder
                 # in image importer kernels.
-                if folder_name != '.logs':
+                if vfolder.name != '.logs':
                     continue
-            if kernel_path_raw := vfolder_mount_map.get(folder_name):
-                if not kernel_path_raw.startswith('/home/work/'):  # type: ignore
-                    raise ValueError(
-                        f'Error while mounting {folder_name} to {kernel_path_raw}: '
-                        'all vfolder mounts should be under /home/work')
-                kernel_path = Path(kernel_path_raw)  # type: ignore
-            else:
-                kernel_path = Path(f'/home/work/{folder_name}')
-            folder_perm = MountPermission(folder_perm_literal)
             mount = Mount(
                 MountTypes.BIND,
-                host_path,
-                kernel_path,
-                folder_perm,
-                is_unmanaged=is_unmanaged,
+                Path(vfolder.host_path),
+                Path(vfolder.kernel_path),
+                vfolder.mount_perm,
             )
             resource_spec.mounts.append(mount)
 
@@ -364,12 +327,11 @@ class AbstractKernelCreationContext(aobject, Generic[KernelObjectType]):
 
         def _mount(
             type, src, dst,
-            is_unmanaged=False,
         ):
             resource_spec.mounts.append(
                 self.get_runner_mount(
                     type, src, dst,
-                    MountPermission('ro'), is_unmanaged=is_unmanaged,
+                    MountPermission('ro'),
                 ),
             )
 
@@ -474,7 +436,7 @@ class AbstractKernelCreationContext(aobject, Generic[KernelObjectType]):
                     if hook_path in already_injected_hooks:
                         continue
                     container_hook_path = f"/opt/kernel/{hook_path.name}"
-                    _mount(MountTypes.BIND, hook_path, container_hook_path, is_unmanaged=True)
+                    _mount(MountTypes.BIND, hook_path, container_hook_path)
                     environ['LD_PRELOAD'] += ':' + container_hook_path
                     already_injected_hooks.add(hook_path)
 
@@ -565,6 +527,7 @@ class AbstractAgent(aobject, Generic[KernelObjectType, KernelCreationContextType
         It must call this super method in an appropriate order, only once.
         """
         self.resource_lock = asyncio.Lock()
+        self.registry_lock = asyncio.Lock()
         self.container_lifecycle_queue = asyncio.Queue()
 
         self.event_producer = await EventProducer.new(
@@ -575,7 +538,6 @@ class AbstractAgent(aobject, Generic[KernelObjectType, KernelCreationContextType
         self.redis_stream_pool = redis.get_redis_object(self.local_config['redis'], db=4)
         self.redis_stat_pool = redis.get_redis_object(self.local_config['redis'], db=0)
 
-        ipc_base_path.mkdir(parents=True, exist_ok=True)
         self.zmq_ctx = zmq.asyncio.Context()
 
         resources_mod.log_alloc_map = self.local_config['debug']['log-alloc-map']
@@ -614,13 +576,14 @@ class AbstractAgent(aobject, Generic[KernelObjectType, KernelCreationContextType
         """
         await cancel_tasks(self._ongoing_exec_batch_tasks)
 
-        # Close all pending kernel runners.
-        for kernel_obj in self.kernel_registry.values():
-            if kernel_obj.runner is not None:
-                await kernel_obj.runner.close()
-            await kernel_obj.close()
-        if stop_signal == signal.SIGTERM:
-            await self.clean_all_kernels(blocking=True)
+        async with self.registry_lock:
+            # Close all pending kernel runners.
+            for kernel_obj in self.kernel_registry.values():
+                if kernel_obj.runner is not None:
+                    await kernel_obj.runner.close()
+                await kernel_obj.close()
+            if stop_signal == signal.SIGTERM:
+                await self.clean_all_kernels(blocking=True)
 
         # Stop timers.
         cancel_results = await cancel_tasks(self.timer_tasks)
@@ -655,10 +618,13 @@ class AbstractAgent(aobject, Generic[KernelObjectType, KernelCreationContextType
         if isinstance(event, KernelTerminatedEvent):
             pending_creation_tasks = self._pending_creation_tasks.get(event.kernel_id, None)
             if pending_creation_tasks is not None:
-                for t in pending_creation_tasks:
-                    if not t.done():
+                for t in set(pending_creation_tasks):
+                    if not t.done() and not t.cancelled():
                         t.cancel()
-                        await t
+                        try:
+                            await t
+                        except asyncio.CancelledError:
+                            continue
         await self.event_producer.produce_event(event, source=self.local_config['agent']['id'])
 
     async def produce_error_event(
@@ -699,6 +665,7 @@ class AbstractAgent(aobject, Generic[KernelObjectType, KernelCreationContextType
                 'images': snappy.compress(msgpack.packb([
                     (repo_tag, digest) for repo_tag, digest in self.images.items()
                 ])),
+                'architecture': get_arch_name(),
             }
             await self.produce_event(AgentHeartbeatEvent(agent_info))
         except asyncio.TimeoutError:
@@ -776,12 +743,13 @@ class AbstractAgent(aobject, Generic[KernelObjectType, KernelCreationContextType
         try:
             updated_kernel_ids = []
             container_ids = []
-            for kernel_id, kernel_obj in [*self.kernel_registry.items()]:
-                if not kernel_obj.stats_enabled:
-                    continue
-                updated_kernel_ids.append(kernel_id)
-                container_ids.append(kernel_obj['container_id'])
-            await self.stat_ctx.collect_container_stat(container_ids)
+            async with self.registry_lock:
+                for kernel_id, kernel_obj in [*self.kernel_registry.items()]:
+                    if not kernel_obj.stats_enabled:
+                        continue
+                    updated_kernel_ids.append(kernel_id)
+                    container_ids.append(kernel_obj['container_id'])
+                await self.stat_ctx.collect_container_stat(container_ids)
             # Let the manager store the statistics in the persistent database.
             if updated_kernel_ids:
                 await self.produce_event(DoSyncKernelStatsEvent(updated_kernel_ids))
@@ -792,54 +760,60 @@ class AbstractAgent(aobject, Generic[KernelObjectType, KernelCreationContextType
             await self.produce_error_event()
 
     async def _handle_start_event(self, ev: ContainerLifecycleEvent) -> None:
-        kernel_obj = self.kernel_registry.get(ev.kernel_id)
-        if kernel_obj is not None:
-            kernel_obj.stats_enabled = True
+        async with self.registry_lock:
+            kernel_obj = self.kernel_registry.get(ev.kernel_id)
+            if kernel_obj is not None:
+                kernel_obj.stats_enabled = True
 
     async def _handle_destroy_event(self, ev: ContainerLifecycleEvent) -> None:
-        result = None
         try:
             current_task = asyncio.current_task()
             assert current_task is not None
             if ev.kernel_id not in self._ongoing_destruction_tasks:
                 self._ongoing_destruction_tasks[ev.kernel_id] = current_task
             self.terminating_kernels.add(ev.kernel_id)
-            kernel_obj = self.kernel_registry.get(ev.kernel_id)
-            if kernel_obj is None:
-                log.warning('destroy_kernel(k:{0}) kernel missing (already dead?)',
-                            ev.kernel_id)
-                if ev.container_id is None:
-                    await self.rescan_resource_usage()
-                    if not ev.suppress_events:
-                        await self.produce_event(
-                            KernelTerminatedEvent(ev.kernel_id, "already-terminated"),
+            async with self.registry_lock:
+                kernel_obj = self.kernel_registry.get(ev.kernel_id)
+                if kernel_obj is None:
+                    log.warning('destroy_kernel(k:{0}) kernel missing (already dead?)',
+                                ev.kernel_id)
+                    if ev.container_id is None:
+                        await self.rescan_resource_usage()
+                        if not ev.suppress_events:
+                            await self.produce_event(
+                                KernelTerminatedEvent(ev.kernel_id, "already-terminated"),
+                            )
+                        if ev.done_future is not None:
+                            ev.done_future.set_result(None)
+                        return
+                    else:
+                        await self.container_lifecycle_queue.put(
+                            ContainerLifecycleEvent(
+                                ev.kernel_id,
+                                ev.container_id,
+                                LifecycleEvent.CLEAN,
+                                ev.reason,
+                                suppress_events=ev.suppress_events,
+                                done_future=ev.done_future,
+                            ),
                         )
-                    return
                 else:
-                    await self.container_lifecycle_queue.put(
-                        ContainerLifecycleEvent(
-                            ev.kernel_id,
-                            ev.container_id,
-                            LifecycleEvent.CLEAN,
-                            ev.reason,
-                            suppress_events=ev.suppress_events,
-                        ),
-                    )
-            else:
-                kernel_obj.stats_enabled = False
-                kernel_obj.termination_reason = ev.reason
-                if kernel_obj.runner is not None:
-                    await kernel_obj.runner.close()
-            result = await self.destroy_kernel(ev.kernel_id, ev.container_id)
+                    kernel_obj.stats_enabled = False
+                    kernel_obj.termination_reason = ev.reason
+                    if kernel_obj.runner is not None:
+                        await kernel_obj.runner.close()
+                    kernel_obj.clean_event = ev.done_future
+                try:
+                    await self.destroy_kernel(ev.kernel_id, ev.container_id)
+                except Exception as e:
+                    if ev.done_future is not None:
+                        ev.done_future.set_exception(e)
+                    raise
         except asyncio.CancelledError:
             pass
         except Exception:
             log.exception('unhandled exception while processing DESTROY event')
             await self.produce_error_event()
-        finally:
-            if ev.done_event is not None:
-                ev.done_event.set()
-                setattr(ev.done_event, '_result', result)
 
     async def _handle_clean_event(self, ev: ContainerLifecycleEvent) -> None:
         destruction_task = self._ongoing_destruction_tasks.get(ev.kernel_id, None)
@@ -847,54 +821,67 @@ class AbstractAgent(aobject, Generic[KernelObjectType, KernelCreationContextType
             # let the destruction task finish first
             await destruction_task
             del destruction_task
-        result = None
-        try:
-            kernel_obj = self.kernel_registry.get(ev.kernel_id)
-            if kernel_obj is not None and kernel_obj.runner is not None:
-                await kernel_obj.runner.close()
-            result = await self.clean_kernel(
-                ev.kernel_id,
-                ev.container_id,
-                ev.kernel_id in self.restarting_kernels,
-            )
-        except Exception:
-            await self.produce_error_event()
-        finally:
+        async with self.registry_lock:
             try:
                 kernel_obj = self.kernel_registry.get(ev.kernel_id)
-                if kernel_obj is not None:
-                    # Restore used ports to the port pool.
-                    port_range = self.local_config['container']['port-range']
-                    # Exclude out-of-range ports, because when the agent restarts
-                    # with a different port range, existing containers' host ports
-                    # may not belong to the new port range.
-                    restored_ports = [*filter(
-                        lambda p: port_range[0] <= p <= port_range[1],
-                        kernel_obj['host_ports'],
-                    )]
-                    self.port_pool.update(restored_ports)
-                    await kernel_obj.close()
-                    # Notify cleanup waiters.
-                    if kernel_obj.clean_event is not None:
-                        kernel_obj.clean_event.set()
-                    # Forget.
-                    self.kernel_registry.pop(ev.kernel_id, None)
+                if kernel_obj is not None and kernel_obj.runner is not None:
+                    await kernel_obj.runner.close()
+                await self.clean_kernel(
+                    ev.kernel_id,
+                    ev.container_id,
+                    ev.kernel_id in self.restarting_kernels,
+                )
+            except Exception as e:
+                if ev.done_future is not None:
+                    ev.done_future.set_exception(e)
+                await self.produce_error_event()
             finally:
-                if ev.done_event is not None:
-                    ev.done_event.set()
-                    setattr(ev.done_event, '_result', result)
-                if restart_tracker := self.restarting_kernels.get(ev.kernel_id, None):
-                    restart_tracker.destroy_event.set()
+                if ev.kernel_id in self.restarting_kernels:
+                    # Don't forget as we are restarting it.
+                    kernel_obj = self.kernel_registry.get(ev.kernel_id, None)
                 else:
-                    await self.rescan_resource_usage()
-                    if not ev.suppress_events:
-                        await self.produce_event(
-                            KernelTerminatedEvent(ev.kernel_id, ev.reason),
-                        )
-                self.terminating_kernels.discard(ev.kernel_id)
+                    # Forget as we are done with this kernel.
+                    kernel_obj = self.kernel_registry.pop(ev.kernel_id, None)
+                try:
+                    if kernel_obj is not None:
+                        # Restore used ports to the port pool.
+                        port_range = self.local_config['container']['port-range']
+                        # Exclude out-of-range ports, because when the agent restarts
+                        # with a different port range, existing containers' host ports
+                        # may not belong to the new port range.
+                        restored_ports = [*filter(
+                            lambda p: port_range[0] <= p <= port_range[1],
+                            kernel_obj['host_ports'],
+                        )]
+                        self.port_pool.update(restored_ports)
+                        await kernel_obj.close()
+                finally:
+                    self.terminating_kernels.discard(ev.kernel_id)
+                    if restart_tracker := self.restarting_kernels.get(ev.kernel_id, None):
+                        restart_tracker.destroy_event.set()
+                    else:
+                        await self.rescan_resource_usage()
+                        if not ev.suppress_events:
+                            await self.produce_event(
+                                KernelTerminatedEvent(ev.kernel_id, ev.reason),
+                            )
+                    # Notify cleanup waiters after all state updates.
+                    if kernel_obj is not None and kernel_obj.clean_event is not None:
+                        kernel_obj.clean_event.set_result(None)
+                    if ev.done_future is not None and not ev.done_future.done():
+                        ev.done_future.set_result(None)
 
     async def process_lifecycle_events(self) -> None:
-        async with aiotools.TaskGroup() as tg:
+
+        async def lifecycle_task_exception_handler(
+            exc_type: Type[Exception], exc_obj: Exception, tb: TracebackType,
+        ) -> None:
+            log.exception("unexpected error in lifecycle task", exc_info=exc_obj)
+
+        async with aiotools.PersistentTaskGroup(
+            exception_handler=lifecycle_task_exception_handler,
+        ) as tg:
+            ipc_base_path = self.local_config['agent']['ipc-base-path']
             while True:
                 ev = await self.container_lifecycle_queue.get()
                 now = time.monotonic()
@@ -933,32 +920,48 @@ class AbstractAgent(aobject, Generic[KernelObjectType, KernelCreationContextType
         *,
         container_id: ContainerId = None,
         exit_code: int = None,
-        done_event: asyncio.Event = None,
-        clean_event: asyncio.Event = None,
+        done_future: asyncio.Future = None,
         suppress_events: bool = False,
     ) -> None:
         try:
             kernel_obj = self.kernel_registry[kernel_id]
             if kernel_obj.termination_reason:
                 reason = kernel_obj.termination_reason
-            if kernel_obj.clean_event is not None:
-                # This should not happen!
-                log.warning('overwriting kernel_obj.clean_event (k:{})', kernel_id)
-            kernel_obj.clean_event = clean_event
-            if container_id is not None and container_id != kernel_obj['container_id']:
-                # This should not happen!
-                log.warning('container id mismatch for kernel_obj (k:{}, c:{}) with event (c:{})',
-                            kernel_id, kernel_obj['container_id'], container_id)
+            if container_id is not None:
+                if event == LifecycleEvent.START:
+                    # Update the container ID (for restarted kernels).
+                    # This will be overwritten by create_kernel() soon, but
+                    # updating here improves consistency of kernel_id to container_id
+                    # mapping earlier.
+                    kernel_obj['container_id'] = container_id
+                elif container_id != kernel_obj['container_id']:
+                    # This should not happen!
+                    log.warning(
+                        "container id mismatch for kernel_obj (k:{}, c:{}) with event (e:{}, c:{})",
+                        kernel_id, kernel_obj['container_id'],
+                        event.name, container_id,
+                    )
             container_id = kernel_obj['container_id']
         except KeyError:
-            pass
+            if event == LifecycleEvent.START:
+                # When creating a new kernel, the kernel_registry is not populated yet
+                # during creation of actual containers.
+                # The Docker daemon may publish the container creation event before
+                # returning the API and our async handlers may deliver the event earlier.
+                # In such cases, it is safe to ignore the missing kernel_regisry item.
+                pass
+            else:
+                log.warning(
+                    "injecting lifecycle event (e:{}) for unknown kernel (k:{})",
+                    event.name, kernel_id,
+                )
         await self.container_lifecycle_queue.put(
             ContainerLifecycleEvent(
                 kernel_id,
                 container_id,
                 event,
                 reason,
-                done_event,
+                done_future,
                 exit_code,
                 suppress_events,
             ),
@@ -1048,19 +1051,19 @@ class AbstractAgent(aobject, Generic[KernelObjectType, KernelCreationContextType
     async def clean_all_kernels(self, blocking: bool = False) -> None:
         kernel_ids = [*self.kernel_registry.keys()]
         clean_events = {}
+        loop = asyncio.get_running_loop()
         if blocking:
             for kernel_id in kernel_ids:
-                clean_events[kernel_id] = asyncio.Event()
+                clean_events[kernel_id] = loop.create_future()
         for kernel_id in kernel_ids:
             await self.inject_container_lifecycle_event(
                 kernel_id,
                 LifecycleEvent.DESTROY,
                 'agent-termination',
-                clean_event=clean_events[kernel_id] if blocking else None,
+                done_future=clean_events[kernel_id] if blocking else None,
             )
         if blocking:
-            waiters = [clean_events[kernel_id].wait()
-                       for kernel_id in kernel_ids]
+            waiters = [clean_events[kernel_id] for kernel_id in kernel_ids]
             await asyncio.gather(*waiters)
 
     @abstractmethod
@@ -1135,6 +1138,7 @@ class AbstractAgent(aobject, Generic[KernelObjectType, KernelCreationContextType
         ``self.kernel_registry`` if any missing.
         """
         try:
+            ipc_base_path = self.local_config['agent']['ipc-base-path']
             with open(ipc_base_path / f'last_registry.{self.local_instance_id}.dat', 'rb') as f:
                 self.kernel_registry = pickle.load(f)
                 for kernel_obj in self.kernel_registry.values():
@@ -1287,6 +1291,13 @@ class AbstractAgent(aobject, Generic[KernelObjectType, KernelCreationContextType
         )
         image_labels = kernel_config['image']['labels']
 
+        agent_architecture = get_arch_name()
+        if agent_architecture != ctx.image_ref.architecture:
+            # disable running different architecture's image
+            raise AgentError(
+                f'cannot run {ctx.image_ref.architecture} image on {agent_architecture} machine',
+            )
+
         # Check if we need to pull the container image
         do_pull = await self.check_image(
             ctx.image_ref,
@@ -1312,9 +1323,10 @@ class AbstractAgent(aobject, Generic[KernelObjectType, KernelCreationContextType
         # adn specify the container mounts.
 
         # Mount backend-specific intrinsic mounts (e.g., scratch directories)
-        resource_spec.mounts.extend(
-            await ctx.get_intrinsic_mounts(),
-        )
+        if not restarting:
+            resource_spec.mounts.extend(
+                await ctx.get_intrinsic_mounts(),
+            )
 
         # Realize ComputeDevice (including accelerators) allocations.
         slots = resource_spec.slots
@@ -1348,15 +1360,18 @@ class AbstractAgent(aobject, Generic[KernelObjectType, KernelCreationContextType
                         raise
 
         # Prepare scratch spaces and dotfiles inside it.
-        await ctx.prepare_scratch()
+        if not restarting:
+            await ctx.prepare_scratch()
 
         # Prepare networking.
         await ctx.apply_network(cluster_info)
         await ctx.install_ssh_keypair(cluster_info)
 
         # Mount vfolders and krunner stuffs.
-        await ctx.mount_vfolders(kernel_config['mounts'], resource_spec)
-        await ctx.mount_krunner(resource_spec, environ)
+        if not restarting:
+            vfolder_mounts = [VFolderMount.from_json(item) for item in kernel_config['mounts']]
+            await ctx.mount_vfolders(vfolder_mounts, resource_spec)
+            await ctx.mount_krunner(resource_spec, environ)
 
         # Inject Backend.AI-intrinsic env-variables for libbaihook and gosu
         label_envs_corecount = image_labels.get('ai.backend.envs.corecount', '')
@@ -1459,26 +1474,35 @@ class AbstractAgent(aobject, Generic[KernelObjectType, KernelCreationContextType
         assert current_task is not None
         self._pending_creation_tasks[kernel_id].add(current_task)
         try:
-            # Wait until bootstrap script is executed.
-            # - Main kernel runner is executed after bootstrap script, and
-            #   check_status is accessible only after kernel runner is loaded.
-            await kernel_obj.check_status()
-
-            # Update the service-ports metadata from the image labels
-            # with the extended template metadata from the agent and krunner.
-            live_services = await kernel_obj.get_service_apps()
-            if live_services['status'] != 'failed':
-                for live_service in live_services['data']:
-                    for service_port in service_ports:
-                        if live_service['name'] == service_port['name']:
-                            service_port.update(live_service)
-                            break
+            async for attempt in AsyncRetrying(
+                wait=wait_fixed(0.3),
+                stop=(stop_after_attempt(10) | stop_after_delay(60)),
+                retry=retry_if_exception_type(zmq.error.ZMQError),
+            ):
+                with attempt:
+                    # Wait until bootstrap script is executed.
+                    # - Main kernel runner is executed after bootstrap script, and
+                    #   check_status is accessible only after kernel runner is loaded.
+                    await kernel_obj.check_status()
+                    # Update the service-ports metadata from the image labels
+                    # with the extended template metadata from the agent and krunner.
+                    live_services = await kernel_obj.get_service_apps()
+                    if live_services['status'] != 'failed':
+                        for live_service in live_services['data']:
+                            for service_port in service_ports:
+                                if live_service['name'] == service_port['name']:
+                                    service_port.update(live_service)
+                                    break
             if self.local_config['debug']['log-kernel-config']:
                 log.debug('service ports:\n{!r}', pretty(service_ports))
-        except (asyncio.CancelledError, zmq.error.ZMQError):
+        except asyncio.CancelledError:
             log.warning("cancelled waiting of container startup (k:{})", kernel_id)
-            raise RuntimeError("cancelled waiting of container startup due to "
-                               "initialization failure or agent shutdown")
+            raise
+        except Exception:
+            log.exception("unexpected error while waiting container startup (k:{})", kernel_id)
+            raise RuntimeError(
+                "cancelled waiting of container startup due to initialization failure",
+            )
         finally:
             self._pending_creation_tasks[kernel_id].remove(current_task)
             if not self._pending_creation_tasks[kernel_id]:
@@ -1659,7 +1683,6 @@ class AbstractAgent(aobject, Generic[KernelObjectType, KernelCreationContextType
                 )
                 raise
             else:
-                # tracker.destroy_event.clear()
                 try:
                     await self.create_kernel(
                         creation_id,
