@@ -23,15 +23,16 @@ agent_local_config_iv = t.Dict({
             tx.HostPortPair(allow_blank_host=True),
         t.Key('agent-sock-port', default=6007): t.Int[1024:65535],
         t.Key('id', default=None): t.Null | t.String,
+        t.Key('ipc-base-path', default="/tmp/backend.ai/ipc"):
+            tx.Path(type='dir', auto_create=True),
         t.Key('region', default=None): t.Null | t.String,
         t.Key('instance-type', default=None): t.Null | t.String,
         t.Key('scaling-group', default='default'): t.String,
-        t.Key('pid-file', default=os.devnull): tx.Path(type='file',
-                                                       allow_nonexisting=True,
-                                                       allow_devnull=True),
+        t.Key('pid-file', default=os.devnull):
+            tx.Path(type='file', allow_nonexisting=True, allow_devnull=True),
         t.Key('event-loop', default='asyncio'): t.Enum('asyncio', 'uvloop'),
         t.Key('skip-manager-detection', default=False): t.ToBool,
-        t.Key('aiomonitor-port', default=50002): t.Int[1:65535],
+        t.Key('aiomonitor-port', default=50102): t.Int[1:65535],
     }).allow_extra('*'),
     t.Key('container'): t.Dict({
         t.Key('kernel-uid', default=-1): tx.UserID,
