@@ -9,6 +9,7 @@ import json
 import logging
 import os
 from pathlib import Path
+from aiohttp import web
 import pkg_resources
 import secrets
 import shutil
@@ -63,7 +64,8 @@ from ai.backend.common.types import (
     current_resource_slots,
 )
 from ai.backend.common.utils import AsyncFileWriter, current_loop
-from .kernel import DockerKernel
+from .kernel import DockerKernel, prepare_kernel_metadata_uri_handling
+from .metadata.server import create_server as create_metadata_server
 from .resources import detect_resources
 from .utils import PersistentServiceContainer
 from ..exception import UnsupportedResource, InitializationError
@@ -78,7 +80,6 @@ from ..agent import (
     AbstractKernelCreationContext,
     ACTIVE_STATUS_SET,
     ComputerContext,
-    ipc_base_path,
 )
 from ..proxy import proxy_connection, DomainSocketProxy
 from ..resources import (
@@ -240,7 +241,9 @@ class DockerKernelCreationContext(AbstractKernelCreationContext[DockerKernel]):
 
         def _create_scratch_dirs():
             self.config_dir.mkdir(parents=True, exist_ok=True)
+            self.config_dir.chmod(0o755)
             self.work_dir.mkdir(parents=True, exist_ok=True)
+            self.work_dir.chmod(0o755)
 
         await loop.run_in_executor(None, _create_scratch_dirs)
 
@@ -357,6 +360,7 @@ class DockerKernelCreationContext(AbstractKernelCreationContext[DockerKernel]):
             Path('/opt/kernel/agent.sock'),
             MountPermission.READ_WRITE,
         ))
+        ipc_base_path = self.local_config['agent']['ipc-base-path']
 
         # domain-socket proxy mount
         # (used for special service containers such image importer)
@@ -393,13 +397,11 @@ class DockerKernelCreationContext(AbstractKernelCreationContext[DockerKernel]):
         src: Union[str, Path],
         target: Union[str, Path],
         perm: Literal['ro', 'rw'] = 'ro',
-        is_unmanaged: bool = False,
         opts: Mapping[str, Any] = None,
     ) -> Mount:
         return Mount(
             type, Path(src), Path(target),
             MountPermission(perm),
-            is_unmanaged=is_unmanaged,
             opts=opts,
         )
 
@@ -775,6 +777,7 @@ class DockerKernelCreationContext(AbstractKernelCreationContext[DockerKernel]):
             agent_config=self.local_config,
             service_ports=service_ports,
             resource_spec=resource_spec,
+            environ=environ,
             data={
                 'container_id': container._id,
                 'kernel_host': advertised_kernel_host or container_bind_host,
@@ -795,6 +798,8 @@ class DockerAgent(AbstractAgent[DockerKernel, DockerKernelCreationContext]):
     agent_sockpath: Path
     agent_sock_task: asyncio.Task
     scan_images_timer: asyncio.Task
+    metadata_server_runner: web.AppRunner
+    docker_ptask_group: aiotools.PersistentTaskGroup
 
     def __init__(
         self,
@@ -823,6 +828,7 @@ class DockerAgent(AbstractAgent[DockerKernel, DockerKernelCreationContext]):
         await self.check_swarm_status()
         if self.heartbeat_extra_info['swarm_enabled']:
             log.info('The Docker Swarm cluster is configured and enabled')
+        ipc_base_path = self.local_config['agent']['ipc-base-path']
         (ipc_base_path / 'container').mkdir(parents=True, exist_ok=True)
         self.agent_sockpath = ipc_base_path / 'container' / f'agent.{self.local_instance_id}.sock'
         socket_relay_name = f"backendai-socket-relay.{self.local_instance_id}"
@@ -837,7 +843,7 @@ class DockerAgent(AbstractAgent[DockerKernel, DockerKernelCreationContext]):
                     'Mounts': [
                         {
                             'Type': 'bind',
-                            'Source': '/tmp/backend.ai/ipc/container',
+                            'Source': str(ipc_base_path / 'container'),
                             'Target': '/ipc',
                         },
                     ],
@@ -850,7 +856,16 @@ class DockerAgent(AbstractAgent[DockerKernel, DockerKernelCreationContext]):
         self.agent_sock_task = asyncio.create_task(self.handle_agent_socket())
         self.monitor_docker_task = asyncio.create_task(self.monitor_docker_events())
         self.monitor_swarm_task = asyncio.create_task(self.check_swarm_status(as_task=True))
+        self.docker_ptask_group = aiotools.PersistentTaskGroup()
 
+        await prepare_kernel_metadata_uri_handling(self.local_config)
+        metadata_server_runner = web.AppRunner(
+            await create_metadata_server(self.local_config, self.kernel_registry),
+        )
+        await metadata_server_runner.setup()
+        site = web.TCPSite(metadata_server_runner, '0.0.0.0', 40128)
+        await site.start()
+        self.metadata_server_runner = metadata_server_runner
         # For legacy accelerator plugins
         self.docker = Docker()
 
@@ -859,6 +874,8 @@ class DockerAgent(AbstractAgent[DockerKernel, DockerKernelCreationContext]):
         if self.agent_sock_task is not None:
             self.agent_sock_task.cancel()
             await self.agent_sock_task
+        if self.docker_ptask_group is not None:
+            await self.docker_ptask_group.shutdown()
 
         try:
             await super().shutdown(stop_signal)
@@ -872,6 +889,7 @@ class DockerAgent(AbstractAgent[DockerKernel, DockerKernelCreationContext]):
             self.monitor_swarm_task.cancel()
             await self.monitor_swarm_task
 
+        await self.metadata_server_runner.cleanup()
         if self.docker:
             await self.docker.close()
 
@@ -1149,7 +1167,7 @@ class DockerAgent(AbstractAgent[DockerKernel, DockerKernelCreationContext]):
                 await self.rescan_resource_usage()
             else:
                 log.exception('destroy_kernel(k:{0}) kill error', kernel_id)
-                self.error_monitor.capture_exception()
+                await self.error_monitor.capture_exception()
 
     async def clean_kernel(
         self,
@@ -1318,13 +1336,19 @@ class DockerAgent(AbstractAgent[DockerKernel, DockerKernelCreationContext]):
                                 log.debug('docker-event: action={}, actor={}',
                                           evdata['Action'], evdata['Actor'])
                             if evdata['Action'] == 'start':
-                                await asyncio.shield(handle_action_start(kernel_id, evdata))
+                                await asyncio.shield(self.docker_ptask_group.create_task(
+                                    handle_action_start(kernel_id, evdata),
+                                ))
                             elif evdata['Action'] == 'die':
-                                await asyncio.shield(handle_action_die(kernel_id, evdata))
+                                await asyncio.shield(self.docker_ptask_group.create_task(
+                                    handle_action_die(kernel_id, evdata),
+                                ))
                         except asyncio.CancelledError:
                             # We are shutting down...
                             return
                         except Exception:
                             log.exception("monitor_docker_events(): unexpected error")
                 finally:
-                    await asyncio.shield(docker.events.stop())
+                    await asyncio.shield(self.docker_ptask_group.create_task(
+                        docker.events.stop(),
+                    ))
