@@ -417,19 +417,23 @@ class MemoryPlugin(AbstractComputePlugin):
             -> Sequence[ContainerMeasurement]:
 
         def get_scratch_size(container_id: str) -> int:
-            for kernel_id, info in ctx.agent.kernel_registry.items():
-                if info['container_id'] == container_id:
-                    break
-            else:
-                return 0
-            work_dir = ctx.agent.local_config['container']['scratch-root'] / str(kernel_id) / 'work'
-            total_size = 0
-            for path in work_dir.rglob('*'):
-                if path.is_symlink():
-                    total_size += path.lstat().st_size
-                elif path.is_file():
-                    total_size += path.stat().st_size
-            return total_size
+            # Temporarily disabled as this function incurs too much delay with
+            # a large number of files in scratch dirs, causing indefinite accumulation of
+            # stat collector tasks and slowing down everything.
+            return 0
+            # for kernel_id, info in ctx.agent.kernel_registry.items():
+            #     if info['container_id'] == container_id:
+            #         break
+            # else:
+            #     return 0
+            # work_dir = ctx.agent.local_config['container']['scratch-root'] / str(kernel_id) / 'work'
+            # total_size = 0
+            # for path in work_dir.rglob('*'):
+            #     if path.is_symlink():
+            #         total_size += path.lstat().st_size
+            #     elif path.is_file():
+            #         total_size += path.stat().st_size
+            # return total_size
 
         async def sysfs_impl(container_id):
             mem_prefix = f'/sys/fs/cgroup/memory/docker/{container_id}/'
