@@ -1,9 +1,11 @@
 import asyncio
 import logging
 import lzma
+import os
 from pathlib import Path, PurePosixPath
 import pkg_resources
 import re
+import shutil
 import subprocess
 import textwrap
 from typing import (
@@ -20,9 +22,11 @@ from aiotools import TaskGroup
 from ai.backend.common.docker import ImageRef
 from ai.backend.common.logging import BraceStyleAdapter
 from ai.backend.common.utils import current_loop
+
+from ai.backend.agent.docker.utils import PersistentServiceContainer
 from ..resources import KernelResourceSpec
 from ..kernel import AbstractKernel, AbstractCodeRunner
-from ..utils import get_arch_name
+from ..utils import closing_async, get_arch_name
 
 log = BraceStyleAdapter(logging.getLogger(__name__))
 
@@ -30,11 +34,13 @@ log = BraceStyleAdapter(logging.getLogger(__name__))
 class DockerKernel(AbstractKernel):
 
     # FIXME: apply TypedDict to data in Python 3.8
+    environ: Mapping[str, Any]
 
     def __init__(self, kernel_id: str, image: ImageRef, version: int, *,
                  agent_config: Mapping[str, Any],
                  resource_spec: KernelResourceSpec,
                  service_ports: Any,  # TODO: type-annotation
+                 environ: Mapping[str, Any],
                  data: Dict[str, Any]) -> None:
         super().__init__(
             kernel_id, image, version,
@@ -42,22 +48,20 @@ class DockerKernel(AbstractKernel):
             resource_spec=resource_spec,
             service_ports=service_ports,
             data=data)
+        self.environ = environ
 
     async def __ainit__(self) -> None:
         await super().__ainit__()
-        self._docker = Docker()
 
     async def close(self) -> None:
-        await self._docker.close()
+        pass
 
     def __getstate__(self):
         props = super().__getstate__()
-        del props['_docker']
         return props
 
     def __setstate__(self, props):
         super().__setstate__(props)
-        self._docker = Docker()
 
     async def create_code_runner(self, *,
                            client_features: FrozenSet[str],
@@ -80,8 +84,9 @@ class DockerKernel(AbstractKernel):
 
     async def get_logs(self):
         container_id = self.data['container_id']
-        container = await self._docker.containers.get(container_id)
-        logs = await container.log(stdout=True, stderr=True)
+        async with closing_async(Docker()) as docker:
+            container = await docker.containers.get(container_id)
+            logs = await container.log(stdout=True, stderr=True)
         return {'logs': ''.join(logs)}
 
     async def interrupt_kernel(self):
@@ -137,23 +142,24 @@ class DockerKernel(AbstractKernel):
 
     async def download_file(self, filepath: str):
         container_id = self.data['container_id']
-        container = self._docker.containers.container(container_id)
-        home_path = PurePosixPath('/home/work')
-        try:
-            abspath = (home_path / filepath)
-            abspath.relative_to(home_path)
-        except ValueError:
-            raise PermissionError('You cannot download files outside /home/work')
-        try:
-            with await container.get_archive(str(abspath)) as tarobj:
-                tarobj.fileobj.seek(0, 2)
-                fsize = tarobj.fileobj.tell()
-                if fsize > 1048576:
-                    raise ValueError('too large file')
-                tarbytes = tarobj.fileobj.getvalue()
-        except DockerError:
-            log.warning('Could not found the file: {0}', abspath)
-            raise FileNotFoundError(f'Could not found the file: {abspath}')
+        async with closing_async(Docker()) as docker:
+            container = docker.containers.container(container_id)
+            home_path = PurePosixPath('/home/work')
+            try:
+                abspath = (home_path / filepath)
+                abspath.relative_to(home_path)
+            except ValueError:
+                raise PermissionError('You cannot download files outside /home/work')
+            try:
+                with await container.get_archive(str(abspath)) as tarobj:
+                    tarobj.fileobj.seek(0, 2)
+                    fsize = tarobj.fileobj.tell()
+                    if fsize > 1048576:
+                        raise ValueError('too large file')
+                    tarbytes = tarobj.fileobj.getvalue()
+            except DockerError:
+                log.warning('Could not found the file: {0}', abspath)
+                raise FileNotFoundError(f'Could not found the file: {abspath}')
         return tarbytes
 
     async def list_files(self, container_path: str):
@@ -339,3 +345,71 @@ async def prepare_krunner_env(local_config: Mapping[str, Any]) -> Mapping[str, S
             continue
         result[distro_name_and_version] = volume_name
     return result
+
+
+LinuxKit_IPTABLES_RULE = \
+    re.compile(r'DNAT\s+tcp\s+\-\-\s+anywhere\s+169\.254\.169\.254\s+tcp dpt:http to:127\.0\.0\.1:50128')
+LinuxKit_CMD_EXEC_PREFIX = [
+    'docker', 'run', '--rm', '-i',
+    '--privileged', '--pid=host',
+    'linuxkit-nsenter:latest',
+]
+
+
+async def prepare_kernel_metadata_uri_handling(local_config: Mapping[str, Any]) -> None:
+    async with closing_async(Docker()) as docker:
+        kernel_version = (await docker.version())['KernelVersion']
+    if 'linuxkit' in kernel_version:
+        local_config['agent']['docker-mode'] = 'linuxkit'
+        # Docker Desktop mode
+        arch = get_arch_name()
+        proxy_worker_binary = pkg_resources.resource_filename(
+            'ai.backend.agent.docker',
+            f'linuxkit-metadata-proxy-worker.{arch}.bin')
+        shutil.copyfile(proxy_worker_binary, '/tmp/backend.ai/linuxkit-metadata-proxy')
+        os.chmod('/tmp/backend.ai/linuxkit-metadata-proxy', 0o755)
+        # Prepare proxy worker container
+        proxy_worker_container = PersistentServiceContainer(
+            'linuxkit-nsenter:latest',
+            {
+                'Cmd': [
+                    '/bin/sh', '-c',
+                    'ctr -n services.linuxkit t kill --exec-id metaproxy docker;'
+                    'ctr -n services.linuxkit t exec --exec-id metaproxy docker '
+                    '/host_mnt/tmp/backend.ai/linuxkit-metadata-proxy -remote-port 40128',
+                ],
+                'HostConfig': {
+                    'PidMode': 'host',
+                    'Privileged': True,
+                },
+            },
+            name='linuxkit-nsenter',
+        )
+        await proxy_worker_container.ensure_running_latest()
+
+        # Check if iptables rule is propagated on LinuxKit VM properly
+        log.info('Checking metadata URL iptables rule ...')
+        proc = await asyncio.create_subprocess_exec(*(
+            LinuxKit_CMD_EXEC_PREFIX +
+            ['/sbin/iptables', '-n', '-t', 'nat', '-L', 'PREROUTING']
+        ), stdout=asyncio.subprocess.PIPE, stderr=asyncio.subprocess.PIPE)
+        await proc.wait()
+        assert proc.stdout is not None
+        raw_rules = await proc.stdout.read()
+        rules = raw_rules.decode()
+        if LinuxKit_IPTABLES_RULE.search(rules) is None:
+            proc = await asyncio.create_subprocess_exec(*(
+                LinuxKit_CMD_EXEC_PREFIX +
+                [
+                    '/sbin/iptables', '-t', 'nat', '-I', 'PREROUTING',
+                    '-d', '169.254.169.254', '-p', 'tcp', '--dport', '80',
+                    '-j', 'DNAT', '--to-destination', '127.0.0.1:50128',
+                ]
+            ))
+            await proc.wait()
+            log.info('Inserted the iptables rules.')
+        else:
+            log.info('The iptables rule already exists.')
+    else:
+        # Linux Mode
+        local_config['agent']['docker-mode'] = 'native'
