@@ -11,6 +11,7 @@ from aiodocker.exceptions import DockerError
 
 from ai.backend.common.logging import BraceStyleAdapter
 
+from ..exception import InitializationError
 from ..utils import closing_async, get_arch_name, update_nested_dict
 
 log = BraceStyleAdapter(logging.getLogger(__name__))
@@ -70,7 +71,7 @@ class PersistentServiceContainer:
                     return 0
                 else:
                     raise
-        return int(img['Config'].get('Labels', {}).get('ai.backend.version', '0'))
+        return int((img['Config'].get('Labels') or {}).get('ai.backend.version', '0'))
 
     async def ensure_running_latest(self) -> None:
         image_version = await self.get_image_version()
@@ -135,7 +136,25 @@ class PersistentServiceContainer:
                 },
             }
             update_nested_dict(container_config, self.container_config)
-            await docker.containers.create(config=container_config, name=self.container_name)
+            try:
+                await docker.containers.create(config=container_config, name=self.container_name)
+            except DockerError as e:
+                err_msg = e.args[1].get("message", "")
+                if (
+                    e.args[0] == 400 and
+                    'bind source path does not exist' in err_msg and
+                    '/tmp/backend.ai/ipc' in err_msg
+                ):
+                    raise InitializationError(
+                        f"Could not create persistent service container '{self.container_name}' "
+                        f"because it cannot access /tmp/backend.ai/ipc directory. "
+                        f"This may occur when Docker is installed with Snap or the agent is configured "
+                        f"to use a private tmp directory. "
+                        f"To resolve, explicitly configure the 'ipc-base-path' option in agent.toml to "
+                        f"indicate a directory under $HOME or a non-virtualized directory.",
+                    )
+                else:
+                    raise
 
     async def start(self) -> None:
         async with closing_async(Docker()) as docker:
