@@ -31,9 +31,10 @@ from async_timeout import timeout
 import zmq, zmq.asyncio
 
 from ai.backend.common import msgpack
+from ai.backend.common.asyncio import current_loop
 from ai.backend.common.docker import ImageRef
+from ai.backend.common.enum_extension import StringSetFlag
 from ai.backend.common.types import aobject, KernelId
-from ai.backend.common.utils import current_loop, StringSetFlag
 from ai.backend.common.logging import BraceStyleAdapter
 from .exception import UnsupportedBaseDistroError
 from .resources import KernelResourceSpec
@@ -140,7 +141,7 @@ class AbstractKernel(UserDict, aobject, metaclass=ABCMeta):
     data: Dict[Any, Any]
     last_used: float
     termination_reason: Optional[str]
-    clean_event: Optional[asyncio.Event]
+    clean_event: Optional[asyncio.Future]
     stats_enabled: bool
 
     _tasks: Set[asyncio.Task]
@@ -295,7 +296,7 @@ class AbstractKernel(UserDict, aobject, metaclass=ABCMeta):
             self._tasks.remove(myself)
 
 
-_zctx = zmq.asyncio.Context()
+_zctx = None
 
 
 class AbstractCodeRunner(aobject, metaclass=ABCMeta):
@@ -324,9 +325,14 @@ class AbstractCodeRunner(aobject, metaclass=ABCMeta):
 
     _closed: bool
 
-    def __init__(self, kernel_id: KernelId, *,
-                 exec_timeout: float = 0,
-                 client_features: FrozenSet[str] = None) -> None:
+    def __init__(
+        self,
+        kernel_id: KernelId,
+        *,
+        exec_timeout: float = 0,
+        client_features: FrozenSet[str] = None,
+    ) -> None:
+        global _zctx
         self.kernel_id = kernel_id
         self.started_at = time.monotonic()
         self.finished_at = None
@@ -336,6 +342,8 @@ class AbstractCodeRunner(aobject, metaclass=ABCMeta):
         self.exec_timeout = exec_timeout
         self.max_record_size = 10 * (2 ** 20)  # 10 MBytes
         self.client_features = client_features or frozenset()
+        if _zctx is None:
+            _zctx = zmq.asyncio.Context()
         self.zctx = _zctx  # share the global context
         self.input_sock = self.zctx.socket(zmq.PUSH)
         self.output_sock = self.zctx.socket(zmq.PULL)
@@ -382,7 +390,10 @@ class AbstractCodeRunner(aobject, metaclass=ABCMeta):
         return props
 
     def __setstate__(self, props):
+        global _zctx
         self.__dict__.update(props)
+        if _zctx is None:
+            _zctx = zmq.asyncio.Context()
         self.zctx = _zctx  # share the global context
         self.input_sock = self.zctx.socket(zmq.PUSH)
         self.output_sock = self.zctx.socket(zmq.PULL)
@@ -560,7 +571,7 @@ class AbstractCodeRunner(aobject, metaclass=ABCMeta):
             await asyncio.sleep(self.exec_timeout)
             if self.output_queue is not None:
                 # TODO: what to do if None?
-                self.output_queue.put(ResultRecord('exec-timeout', None))
+                await self.output_queue.put(ResultRecord('exec-timeout', None))
         except asyncio.CancelledError:
             pass
 
