@@ -2,6 +2,8 @@ import functools
 from io import StringIO
 from pathlib import Path
 import uuid
+
+import aiotools
 from ai.backend.common.asyncio import current_loop
 from ai.backend.common.docker import ImageRef
 import asyncio
@@ -13,8 +15,16 @@ import random
 import shutil
 import signal
 from typing import (
-    Any, FrozenSet, List, Literal, Mapping,
-    MutableMapping, Optional, Sequence, Tuple, Union, cast,
+    Any,
+    FrozenSet,
+    List,
+    Literal,
+    Mapping,
+    MutableMapping,
+    Optional,
+    Sequence,
+    Tuple,
+    Union,
 )
 from ai.backend.common.logging_utils import BraceStyleAdapter
 
@@ -34,7 +44,6 @@ from .kube_object import (
 )
 
 from ..agent import ACTIVE_STATUS_SET, AbstractAgent, AbstractKernelCreationContext, ComputerContext
-from ..defs import ipc_base_path
 from ..exception import K8sError, UnsupportedResource
 from ..kernel import KernelFeatures
 from ..resources import (
@@ -56,12 +65,10 @@ from ai.backend.common.types import (
     KernelCreationConfig,
     KernelId,
     MountPermission,
-    MountTuple3,
-    MountTuple4,
-    MountTuple5,
     MountTypes,
     ResourceSlot,
     SlotName,
+    VFolderMount,
     current_resource_slots,
 )
 
@@ -255,7 +262,7 @@ class KubernetesKernelCreationContext(AbstractKernelCreationContext[KubernetesKe
 
             self.config_maps.append(cm)
 
-        self.process_volumes([
+        await self.process_volumes([
             KubernetesConfigMapVolume(
                 name=f'kernel-{self.kernel_id}-ssh-keypair',
                 configMap={
@@ -263,7 +270,7 @@ class KubernetesKernelCreationContext(AbstractKernelCreationContext[KubernetesKe
                 },
             ),
         ])
-        self.process_mounts([
+        await self.process_mounts([
             Mount(
                 MountTypes.K8S_GENERIC,
                 Path('public'),
@@ -323,17 +330,18 @@ class KubernetesKernelCreationContext(AbstractKernelCreationContext[KubernetesKe
         src: Union[str, Path],
         target: Union[str, Path],
         perm: Literal['ro', 'rw'] = 'ro',
-        is_unmanaged: bool = False,
         opts: Mapping[str, Any] = None,
     ) -> Mount:
         return Mount(
-                MountTypes.K8S_GENERIC, Path(src), Path(target),
-                MountPermission(perm),
-                opts={
-                        **(opts or {}),
-                        'name': f'kernel-{self.kernel_id}-scratches',
-                     },
-               )
+            MountTypes.K8S_GENERIC,
+            Path(src),
+            Path(target),
+            MountPermission(perm),
+            opts={
+                **(opts or {}),
+                'name': f'kernel-{self.kernel_id}-scratches',
+            },
+        )
 
     async def process_volumes(
         self,
@@ -343,76 +351,35 @@ class KubernetesKernelCreationContext(AbstractKernelCreationContext[KubernetesKe
 
     async def mount_vfolders(
         self,
-        vfolders,
+        vfolders: Sequence[VFolderMount],
         resource_spec: KernelResourceSpec,
     ) -> None:
-        vfolder_mount_map: Mapping[str, str]
-        vfolder_mount_map = self.kernel_config.get('mount_map', {})
-
         # We can't mount vFolder backed by storage proxy
-        for i, vfolder in zip(range(len(vfolders)), vfolders):
-            if len(vfolder) == 5:
-                folder_name, folder_host, folder_id, folder_perm_literal, host_path_raw = \
-                        cast(MountTuple5, vfolder)
-                if host_path_raw:
-                    host_path = Path(host_path_raw)
-                else:
-                    mount_path = Path(folder_id)
-                    if mount_path.is_absolute():
-                        # Use the storage proxy-provided path as-is.
-                        host_path = mount_path
-                    else:
-                        host_path = (Path(folder_host) /
-                                     self.local_config['vfolder']['fsprefix'] / folder_id)
-            elif len(vfolder) == 4:  # for backward compatibility
-                folder_name, folder_host, folder_id, folder_perm_literal = \
-                    cast(MountTuple4, vfolder)
-                host_path = (Path(folder_host) /
-                             self.local_config['vfolder']['fsprefix'] / folder_id)
-            elif len(vfolder) == 3:  # legacy managers
-                folder_name, folder_host, folder_id = cast(MountTuple3, vfolder)
-                folder_perm_literal = 'rw'
-                host_path = (Path(folder_host) /
-                             self.local_config['vfolder']['fsprefix'] / folder_id)
-            else:
-                raise RuntimeError(
-                    'Unexpected number of vfolder mount detail tuple size')
-
+        for idx, vfolder in enumerate(vfolders):
             if self.internal_data.get('prevent_vfolder_mounts', False):
                 # Only allow mount of ".logs" directory to prevent expose
                 # internal-only information, such as Docker credentials to user's ".docker" vfolder
                 # in image importer kernels.
-                if folder_name != '.logs':
+                if vfolder.name != '.logs':
                     continue
-            if kernel_path_raw := vfolder_mount_map.get(folder_name):
-                if not kernel_path_raw.startswith('/home/work/'):  # type: ignore
-                    raise ValueError(
-                        f'Error while mounting {folder_name} to {kernel_path_raw}: '
-                        'all vfolder mounts should be under /home/work')
-                kernel_path = Path(kernel_path_raw)  # type: ignore
-            else:
-                kernel_path = Path(f'/home/work/{folder_name}')
-            folder_perm = MountPermission(folder_perm_literal)
-
             mount = Mount(
                 MountTypes.K8S_HOSTPATH,
-                None,
-                kernel_path,
-                folder_perm,
+                Path(vfolder.host_path),
+                Path(vfolder.kernel_path),
+                vfolder.mount_perm,
                 opts={
-                    'name': f'kernel-{self.kernel_id}-hostPath-{i}',
+                    'name': f'kernel-{self.kernel_id}-hostPath-{idx}',
                 },
             )
-            self.process_volumes([
+            await self.process_volumes([
                 KubernetesHostPathVolume(
-                    name=f'kernel-{self.kernel_id}-hostPath-{i}',
+                    name=f'kernel-{self.kernel_id}-hostPath-{idx}',
                     hostPath={
-                        'path': host_path.as_posix(),
+                        'path': vfolder.host_path.as_posix(),
                         'type': 'Directory',
                     },
                 ),
             ])
-
             resource_spec.mounts.append(mount)
 
     async def apply_accelerator_allocation(self, computer, device_alloc) -> None:
@@ -533,7 +500,7 @@ class KubernetesKernelCreationContext(AbstractKernelCreationContext[KubernetesKe
                 ssh_config_map = ConfigMap(self.kernel_id, f'kernel-{self.kernel_id}-ssh-config')
                 ssh_config_map.put('authorized_keys', pubkey)
                 ssh_config_map.put('id_container', privkey)
-                self.process_volumes([
+                await self.process_volumes([
                     KubernetesConfigMapVolume(
                         name='ssh-config',
                         configMap={
@@ -541,7 +508,7 @@ class KubernetesKernelCreationContext(AbstractKernelCreationContext[KubernetesKe
                         },
                     ),
                 ])
-                self.process_mounts([
+                await self.process_mounts([
                     Mount(
                         MountTypes.K8S_GENERIC,
                         Path('authorized_keys'),
@@ -710,6 +677,7 @@ class KubernetesKernelCreationContext(AbstractKernelCreationContext[KubernetesKe
 
 class KubernetesAgent(AbstractAgent[KubernetesKernel, KubernetesKernelCreationContext]):
     workers: MutableMapping[str, MutableMapping[str, str]] = {}
+    k8s_ptask_group: aiotools.PersistentTaskGroup
 
     def __init__(
         self,
@@ -730,11 +698,12 @@ class KubernetesAgent(AbstractAgent[KubernetesKernel, KubernetesKernelCreationCo
 
     async def __ainit__(self) -> None:
         await super().__ainit__()
-        (ipc_base_path / 'container').mkdir(parents=True, exist_ok=True)
+        ipc_base_path = self.local_config['agent']['ipc-base-path']
         self.agent_sockpath = ipc_base_path / 'container' / f'agent.{self.local_instance_id}.sock'
 
         await self.check_krunner_pv_status()
         await self.fetch_workers()
+        self.k8s_ptask_group = aiotools.PersistentTaskGroup()
         # Socket Relay initialization
         # Agent socket handler initialization
         # K8s event monitor task initialization
@@ -830,6 +799,8 @@ class KubernetesAgent(AbstractAgent[KubernetesKernel, KubernetesKernelCreationCo
         # Stop agent socket handler task
 
         try:
+            if self.k8s_ptask_group is not None:
+                await self.k8s_ptask_group.shutdown()
             await super().shutdown(stop_signal)
         finally:
             # Stop k8s event monitoring.
@@ -928,8 +899,7 @@ class KubernetesAgent(AbstractAgent[KubernetesKernel, KubernetesKernelCreationCo
         except:
             log.warning('_destroy_kernel({0}) kernel missing (already dead?)',
                         kernel_id)
-
-            await asyncio.shield(force_cleanup())
+            await asyncio.shield(self.k8s_ptask_group.create_task(force_cleanup()))
             return None
         deployment_name = kernel['deployment_name']
         try:
