@@ -58,11 +58,16 @@ if TYPE_CHECKING:
 log = BraceStyleAdapter(logging.getLogger('ai.backend.agent.resources'))
 
 log_alloc_map: bool = False
-
-
 known_slot_types: Mapping[SlotName, SlotTypes] = {}
 
 
+def round_down(from_dec: Decimal, with_dec: Decimal):
+    remainder = from_dec.remainder_near(with_dec)
+    if remainder < 0:
+        remainder += with_dec
+    return from_dec - remainder
+
+
 class AllocationStrategy(enum.Enum):
     FILL = 0
     EVENLY = 1
@@ -378,7 +383,6 @@ class Mount:
     target: Path
     permission: MountPermission = MountPermission.READ_ONLY
     opts: Optional[Mapping[str, Any]] = None
-    is_unmanaged: bool = False
 
     def __str__(self):
         return f'{self.source}:{self.target}:{self.permission.value}'
@@ -735,19 +739,27 @@ class FractionAllocMap(AbstractAllocMap):
                         f"Slots {slot_name_a} and {slot_name_b} cannot be allocated at the same time.",
                     )
 
-        # check quantum size
-        for slot_name, slot_value in requested_slots.items():
-            if slot_value.remainder_near(self.quantum_size) != 0:
-                raise NotMultipleOfQuantum(
-                    f"Requested amount {slot_value} for {slot_name} is "
-                    f"not a multiple of {self.quantum_size}.",
-                )
-
-        return self._allocate_impl[self.allocation_strategy](
+        calculated_alloc_map = self._allocate_impl[self.allocation_strategy](
             requested_slots,
             context_tag=context_tag,
             min_memory=min_memory,
         )
+        actual_alloc_map: MutableMapping[SlotName, MutableMapping[DeviceId, Decimal]] = {}
+        for slot_name, alloc in calculated_alloc_map.items():
+            actual_alloc: MutableMapping[DeviceId, Decimal] = {}
+            for dev_id, val in alloc.items():
+                self.allocations[slot_name][dev_id] = round_down(
+                    self.allocations[slot_name][dev_id], self.quantum_size)
+                actual_alloc[dev_id] = round_down(val, self.quantum_size)
+            if sum(actual_alloc.values()) == 0 and requested_slots[slot_name] > 0:
+                raise NotMultipleOfQuantum(
+                    f'Requested resource amount for {slot_name} is {requested_slots[slot_name]} '
+                    'but actual calculated amount is zero. This can happen if user requests '
+                    'resource amount smaller than target device\'s quantum size.',
+                )
+            actual_alloc_map[slot_name] = actual_alloc
+
+        return actual_alloc_map
 
     def _allocate_by_filling(
         self,
@@ -802,11 +814,7 @@ class FractionAllocMap(AbstractAllocMap):
                     remaining_alloc -= allocated
                 if remaining_alloc <= 0:
                     break
-            if any(value.remainder_near(self.quantum_size) != 0 for value in slot_allocation.values()):
-                alloc_repr = ", ".join(f"{k}={v}" for k, v in slot_allocation.items())
-                raise InsufficientResource(
-                    f"Device allocation ({alloc_repr}) is not a multiple of {self.quantum_size}",
-                )
+
             allocation[slot_name] = slot_allocation
         return allocation
 
@@ -972,11 +980,6 @@ class FractionAllocMap(AbstractAllocMap):
                 slot_allocation = sorted(best_alloc_candidate_arr,
                                          key=operator.itemgetter(1, 2, 3))[-1][0]
             allocation[slot_name] = slot_allocation
-            if any(value.remainder_near(self.quantum_size) != 0 for value in slot_allocation.values()):
-                alloc_repr = ", ".join(f"{k}={v}" for k, v in slot_allocation.items())
-                raise InsufficientResource(
-                    f"Device allocation ({alloc_repr}) is not a multiple of {self.quantum_size}",
-                )
             for dev_id, value in slot_allocation.items():
                 self.allocations[slot_name][dev_id] += value
         return allocation
