@@ -8,6 +8,7 @@ import logging, logging.config
 import os, os.path
 from pathlib import Path
 from pprint import pformat, pprint
+import shutil
 import signal
 import sys
 from typing import (
@@ -33,7 +34,9 @@ from callosum.rpc import Peer, RPCMessage
 from callosum.ordering import ExitOrderedAsyncScheduler
 from callosum.lower.zeromq import ZeroMQAddress, ZeroMQRPCTransport
 import click
+from etcetra.types import WatchEventType
 from setproctitle import setproctitle
+import tomlkit
 from trafaret.dataerror import DataError as TrafaretDataError
 
 from ai.backend.common import config, utils, identity, msgpack
@@ -176,8 +179,6 @@ class AgentRPCServer(aobject):
 
     async def __ainit__(self) -> None:
         # Start serving requests.
-        self._create_sema = asyncio.Semaphore(4)
-        self._destroy_sema = asyncio.Semaphore(4)
         await self.update_status('starting')
 
         if not self.skip_detect_manager:
@@ -227,7 +228,7 @@ class AgentRPCServer(aobject):
             log.warning('watching etcd to wait for the manager being available')
             async with aclosing(self.etcd.watch_prefix('nodes/manager')) as agen:
                 async for ev in agen:
-                    if ev.event == 'put' and ev.value == 'up':
+                    if ev.event == WatchEventType.PUT and ev.value == 'up':
                         break
         log.info('detected at least one manager running')
 
@@ -285,6 +286,19 @@ class AgentRPCServer(aobject):
     async def update_status(self, status):
         await self.etcd.put('', status, scope=ConfigScopes.NODE)
 
+    @rpc_function
+    @collect_error
+    async def update_scaling_group(self, scaling_group):
+        cfg_src_path = config.find_config_file('agent')
+        with open(cfg_src_path, 'r') as f:
+            data = tomlkit.load(f)
+            data['agent']['scaling-group'] = scaling_group
+        shutil.copy(cfg_src_path, f"{cfg_src_path}.bak")
+        with open(cfg_src_path, 'w') as f:
+            tomlkit.dump(data, f)
+        self.local_config['agent']['scaling-group'] = scaling_group
+        log.info('rpc::update_scaling_group()')
+
     @rpc_function
     @collect_error
     async def ping(self, msg: str) -> str:
@@ -312,30 +326,29 @@ class AgentRPCServer(aobject):
         raw_configs: Sequence[dict],
         raw_cluster_info: dict,
     ):
-        async with self._create_sema:
-            cluster_info = cast(ClusterInfo, raw_cluster_info)
-            session_id = SessionId(UUID(raw_session_id))
-            raw_results = []
-            coros = []
-            for raw_kernel_id, raw_config in zip(raw_kernel_ids, raw_configs):
-                log.info('rpc::create_kernel(k:{0}, img:{1})',
-                        raw_kernel_id, raw_config['image']['canonical'])
-                kernel_id = KernelId(UUID(raw_kernel_id))
-                kernel_config = cast(KernelCreationConfig, raw_config)
-                coros.append(self.agent.create_kernel(
-                    creation_id,
-                    session_id,
-                    kernel_id,
-                    kernel_config,
-                    cluster_info,
-                ))
-            results = await asyncio.gather(*coros, return_exceptions=True)
+        cluster_info = cast(ClusterInfo, raw_cluster_info)
+        session_id = SessionId(UUID(raw_session_id))
+        raw_results = []
+        coros = []
+        for raw_kernel_id, raw_config in zip(raw_kernel_ids, raw_configs):
+            log.info('rpc::create_kernel(k:{0}, img:{1})',
+                     raw_kernel_id, raw_config['image']['canonical'])
+            kernel_id = KernelId(UUID(raw_kernel_id))
+            kernel_config = cast(KernelCreationConfig, raw_config)
+            coros.append(self.agent.create_kernel(
+                creation_id,
+                session_id,
+                kernel_id,
+                kernel_config,
+                cluster_info,
+            ))
+        results = await asyncio.gather(*coros, return_exceptions=True)
         errors = [*filter(lambda item: isinstance(item, Exception), results)]
         if errors:
             # Raise up the first error.
             if len(errors) == 1:
                 raise errors[0]
-            raise aiotools.MultiError("agent.create_kernels() failed", errors=errors)
+            raise aiotools.TaskGroupError("agent.create_kernels() failed", errors)
         raw_results = [
             {
                 'id': str(result['id']),
@@ -361,18 +374,17 @@ class AgentRPCServer(aobject):
         reason: str = None,
         suppress_events: bool = False,
     ):
-        async with self._destroy_sema:
-            log.info('rpc::destroy_kernel(k:{0})', kernel_id)
-            done = asyncio.Event()
-            await self.agent.inject_container_lifecycle_event(
-                KernelId(UUID(kernel_id)),
-                LifecycleEvent.DESTROY,
-                reason or 'user-requested',
-                done_event=done,
-                suppress_events=suppress_events,
-            )
-            await done.wait()
-        return getattr(done, '_result', None)
+        loop = asyncio.get_running_loop()
+        done = loop.create_future()
+        log.info('rpc::destroy_kernel(k:{0})', kernel_id)
+        await self.agent.inject_container_lifecycle_event(
+            KernelId(UUID(kernel_id)),
+            LifecycleEvent.DESTROY,
+            reason or 'user-requested',
+            done_future=done,
+            suppress_events=suppress_events,
+        )
+        return await done
 
     @rpc_function
     @collect_error
@@ -594,9 +606,10 @@ async def server_main(
 
     rpc_addr = local_config['agent']['rpc-listen-addr']
     if not rpc_addr.host:
-        subnet_hint = await etcd.get('config/network/subnet/agent')
-        if subnet_hint is not None:
-            subnet_hint = ip_network(subnet_hint)
+        _subnet_hint = await etcd.get('config/network/subnet/agent')
+        subnet_hint = None
+        if _subnet_hint is not None:
+            subnet_hint = ip_network(_subnet_hint)
         log.debug('auto-detecting agent host')
         local_config['agent']['rpc-listen-addr'] = HostPortPair(
             await identity.get_instance_ip(subnet_hint),
@@ -764,6 +777,7 @@ def main(
                     server_main_logwrapper,
                     num_workers=1,
                     args=(cfg, log_endpoint),
+                    wait_timeout=5.0,
                 )
                 log.info('exit.')
         finally:
