@@ -10,7 +10,7 @@ import json
 import logging
 from pathlib import Path
 from typing import (
-    Any, Optional, Union,
+    Any, Collection, Optional, Union,
     TypedDict,
     Mapping, MutableMapping, Dict,
     Sequence, List, Tuple,
@@ -77,11 +77,11 @@ class ServiceParser:
     async def start_service(
         self,
         service_name: str,
-        frozen_envs: List[str],
+        frozen_envs: Collection[str],
         opts: Mapping[str, Any],
-    ) -> Tuple[Optional[Sequence[str]], Optional[Mapping[str, str]]]:
+    ) -> Tuple[Optional[Sequence[str]], Mapping[str, str]]:
         if service_name not in self.services.keys():
-            return None, None
+            return None, {}
         service = self.services[service_name]
         if service.noop:
             return [], {}
@@ -102,7 +102,7 @@ class ServiceParser:
             cmdargs.append(arg.format_map(self.variables))
 
         additional_arguments = dict(service.default_arguments)
-        if 'arguments' in opts.keys():
+        if 'arguments' in opts.keys() and opts['arguments']:
             for argname, argvalue in opts['arguments'].items():
                 if argname not in service.allowed_arguments:
                     raise DisallowedArgument(
@@ -112,7 +112,7 @@ class ServiceParser:
         for env_name, env_value in service.env.items():
             env[env_name.format_map(self.variables)] = env_value.format_map(self.variables)
 
-        if 'envs' in opts.keys():
+        if 'envs' in opts.keys() and opts['envs']:
             for envname, envvalue in opts['envs'].items():
                 if envname not in service.allowed_envs:
                     raise DisallowedEnvironment(
